#include <cassert>

#include "shader.h"
#include "shader_vars.h"

#include <sstream>
#include <iomanip>

namespace shader {

  using namespace std;

  refcounted::~refcounted ()
  {
    if (refcnt)
      abort ();
  }

  void refcounted::refcnt_destroy () const
  {
    if (!--refcnt)
      delete this; // quite a bit of code...
  }

  const char str_float  [] = "float";
  const char str_vec2   [] = "vec2";
  const char str_vec3   [] = "vec3";
  const char str_vec4   [] = "vec4";
  const char str_mat2   [] = "mat2";
  const char str_mat3   [] = "mat3";
  const char str_mat4   [] = "mat4";

  const char str_sampler_1d             [] = "sampler1D";
  const char str_sampler_1d_shadow      [] = "sampler1DShadow";
  const char str_sampler_2d             [] = "sampler2D";
  const char str_sampler_2d_shadow      [] = "sampler2DShadow";
  const char str_sampler_2d_rect        [] = "sampler2DRect";
  const char str_sampler_2d_rect_shadow [] = "sampler2DRectShadow";
  const char str_sampler_3d             [] = "sampler3D";
  const char str_sampler_3d_rect        [] = "sampler3DRect";
  const char str_sampler_cube           [] = "samplerCube";

  unsigned int var_i::next_id = 0;

  var_i::var_i (const char *domainstr, const char *typestr)
  : domainstr (domainstr), typestr (typestr)
  {
  }

  var_i::~var_i ()
  {
  }

#if 0
  stream_i::stream_i (const char *strtype)
  : var_i (strtype)
  {
    sprintf (name, "V%d", ++next_id);
  }
#endif

  temporary_i::temporary_i (const char *strtype)
  : var_i (0, strtype)
  {
    sprintf (name, "T%d", ++next_id);
  }

  varying_i::varying_i (const char *strtype)
  : var_i ("varying", strtype)
  {
    sprintf (name, "V%d", ++next_id);
  }

  uniform_i::uniform_i (const char *strtype)
  : var_i ("uniform", strtype)
  {
    sprintf (name, "U%d", ++next_id);
  }

  ////////////////////////////////////////////////////////////////////////////

  void var_i::operator ()() const
  {
    shader_builder::cur->code << name;

    if (shader_builder::cur->first (this))
      {
        if (domainstr)
          shader_builder::cur->global << domainstr << ' ' << typestr << ' ' << name << ";\n";
        else
          shader_builder::cur->local << "  " << typestr << ' ' << name << ";\n";
      }
  }

  ////////////////////////////////////////////////////////////////////////////

  int texture_units::unit_count = 8;
  int texture_units::units[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };

  shader_object_i *cur = 0;

  shader_object_i::shader_object_i (GLenum type)
  : type (type)
  {
    id = glCreateShaderObjectARB (type);
    assert (id);
  }

  shader_object_i::~shader_object_i ()
  {
    glDeleteObjectARB (id);
  }

  static string linify (const string &s)
  {
    ostringstream o;

    int b = 0, e;
    int l = 1;
    do {
      o << setw (3) << l << ": ";
      e = s.find ('\n', b);
      if (e == string::npos)
        e = s.size ();

      o << s.substr (b, e - b + 1);
      b = e + 1;
      l++;
    } while (b < s.size ());

    return o.str ();
  }

  ////////////////////////////////////////////////////////////////////////////
  
  GLint uniform_i::location ()
  {
    assert (program_object_i::cur);

    GLint &rid = program_object_i::cur->uloc[this];

    if (!rid)
      rid = glGetUniformLocationARB (program_object_i::cur->id, name);

    return rid;
  }

  program_object_i *program_object_i::cur;

  program_object_i::program_object_i ()
  {
    id = glCreateProgramObjectARB ();
    assert (id);
  }

  program_object_i::~program_object_i ()
  {
    glDeleteProgramsARB (1, &id);
  }

  void program_object_i::link ()
  {
    glLinkProgramARB (id);
        
    GLint linked;
    glGetObjectParameterivARB (id, GL_OBJECT_LINK_STATUS_ARB, &linked);

    if (!linked)
      {
        char infolog[8192];
        glGetInfoLogARB (id, 8192, NULL, infolog);
        printf ("LINK-INFOLOG<%s>\n", infolog);
        abort ();
      }

    uloc.clear ();
  }

  void program_object_i::enable ()
  {
    if (this != cur)
      {
        glUseProgramObjectARB (id);
        //TODO: set samplers here?
        cur = this;
      }
  }

  void program_object_i::disable ()
  {
    //TODO: clear samplers here?
    glUseProgramObjectARB (0);
    cur = 0;
  }

  static map<string, program_object> progcache;

  program_object get_program (const string &vsh, const string &fsh)
  {
    string idx = vsh + "\0" + fsh;

    map<string, program_object>::iterator i = progcache.find (idx);

    if (i != progcache.end ())
      return i->second;

    program_object p;

    if (vsh.size ())
      {
        vertex_shader sh;
        p->attach (sh);
        sh->compile (vsh);
      }

    if (fsh.size ())
      {
        fragment_shader sh;
        p->attach (sh);
        sh->compile (fsh);
      }

    p->link ();

    progcache.insert (pair<string, program_object> (idx, p));

    return p;
  }

  const sl_expr< sl_string<60> > sl_convert< ::vec2 >::convert (const ::vec2 &v)
  {
    sl_string<60> s;
    sprintf (s.str, "vec2 (%e, %e)", v.x, v.y);
    return s;
  }

  const sl_expr< sl_string<80> > sl_convert< ::vec3 >::convert (const ::vec3 &v)
  {
    sl_string<80> s;
    sprintf (s.str, "vec3 (%e, %e, %e)", v.x, v.y, v.z);
    return s;
  }

  const sl_expr< sl_string<100> > sl_convert< ::vec4 >::convert (const ::vec4 &v)
  {
    sl_string<100> s;
    sprintf (s.str, "vec4 (%e, %e, %e, %e)", v.x, v.y, v.z, v.w);
    return s;
  }

  shader_builder *shader_builder::cur = 0;

  bool shader_builder::first (const void *p)
  {
    if (seen.find (p) == seen.end ())
      {
        seen.insert (p);
        return true;
      }

    return false;
  }

  void shader_builder::start ()
  {
    cur = new shader_builder;
    cur->code << scientific;
  }

  string shader_builder::stop ()
  {
    ostringstream os;

    os << cur->global.str ()
       << "\nvoid main (void)\n"
       << "{\n"
       << cur->local.str ()
       << "\n"
       << cur->code.str ()
       << "}\n";

    delete cur;
    cur = 0;

    return os.str ();
  }

  void shader_object_i::compile (const string &source)
  {
    const char *sptr = source.data ();
    const int slen = source.size ();

    printf ("%s\n", linify (source).c_str ());
    glShaderSourceARB (id, 1, &sptr, &slen);
    glCompileShaderARB (id);

    GLint compiled;
    glGetObjectParameterivARB (id, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);

    if (!compiled)
      {
        char infolog[8192];
        glGetInfoLogARB (id, 8192, NULL, infolog);
        printf ("%s\n", linify (source).c_str ());
        printf ("%s\n", infolog);
        abort ();
      }
  }

}

