#ifndef UTIL_H
#define UTIL_H

#include <cmath>
#include <cfloat>
#include <cstdlib>
#include <vector>
#include <string>

#include "opengl.h"

#include <SDL/SDL_image.h>

using namespace std;

typedef long long soffs; // 32 bit
typedef unsigned long long uoffs;
#define OFFS_BITS 63
#define SOFFS_MIN (soffs)-(1LL << (OFFS_BITS - 2))
#define SOFFS_MAX (soffs)+(1LL << (OFFS_BITS - 2))
#define MAXEXTENT (1ULL << (OFFS_BITS - 1))

#define ABS(n) ((n) < 0 ? -(n) : (n))

struct sector
{
  soffs x, y, z;

  sector (soffs x, soffs y, soffs z) : x(x), y(y), z(z) { };
  sector (soffs xyz = 0) : x(xyz), y(xyz), z(xyz) { };
};

inline sector operator +(const sector &a, const sector &b)
{
  return sector (a.x + b.x, a.y + b.y, a.z + b.z);
}

inline sector operator -(const sector &a, const sector &b)
{
  return sector (a.x - b.x, a.y - b.y, a.z - b.z);
}

inline sector operator /(const sector &a, soffs b)
{
  return sector (a.x / b, a.y / b, a.z / b);
}

inline sector operator >>(const sector &a, unsigned int b)
{
  return sector (a.x >> b, a.y >> b, a.z >> b);
}

inline bool operator <=(const sector &a, const sector &b)
{
  return a.x <= b.x && a.y <= b.y && a.z <= b.z;
}

inline soffs max (const sector &a)
{
  return max (a.x, max (a.y, a.z));
}

inline sector translate (const sector &p, const sector &src, const sector &dst)
{
  return p + (dst - src);
}

inline sector abs (const sector &s)
{
  return sector (ABS (s.x), ABS (s.y), ABS (s.z));
}

struct vec2
{
  GLfloat x, y;
  vec2 () { };
  vec2 (GLfloat s) : x(s), y(s) { };
  vec2 (GLfloat x, GLfloat y) : x(x), y(y) { };
};

inline const vec2 operator +(const vec2 &a, const vec2 &b)
{
  return vec2 (a.x + b.x, a.y + b.y);
}

inline const vec2 operator -(const vec2 &a, const vec2 &b)
{
  return vec2 (a.x - b.x, a.y - b.y);
}

struct vec3
{
  GLfloat x, y, z;
  vec3 () { };
  vec3 (GLfloat s) : x(s), y(s), z(s) { };
  vec3 (GLfloat x, GLfloat y, GLfloat z) : x(x), y(y), z(z) { };
  vec3 (const sector &s) : x(s.x), y(s.y), z(s.z) { };

  const vec3 operator -() const
  { return vec3 (-x, -y, -z); }
};

const vec3 normalize (const vec3 &v);
const vec3 cross (const vec3 &a, const vec3 &b);

inline const vec3 operator *(const vec3 &a, GLfloat s)
{
  return vec3 (a.x * s, a.y * s, a.z * s);
}

inline const vec3 operator +(const vec3 &a, const vec3 &b)
{
  return vec3 (a.x + b.x, a.y + b.y, a.z + b.z);
}

inline const vec3 operator -(const vec3 &a, const vec3 &b)
{
  return vec3 (a.x - b.x, a.y - b.y, a.z - b.z);
}

inline GLfloat dot (const vec3 &a, const vec3 &b)
{
  return a.x * b.x + a.y * b.y + a.z * b.z;
}

// squared length
inline const GLfloat length2 (const vec3 &v)
{
  return dot (v, v);
}

inline const GLfloat length (const vec3 &v)
{
  return sqrtf (length2 (v));
}

typedef vec3 point;

struct vec4
{
  GLfloat x, y, z, w;
  vec4 () { };
  vec4 (GLfloat s) : x(s), y(s), z(s), w(s) { };
  vec4 (GLfloat x, GLfloat y, GLfloat z, GLfloat w) : x(x), y(y), z(z), w(w) { };
  vec4 (const vec3 &v, GLfloat w = 1.F) : x(v.x), y(v.y), z(v.z), w(w) { };
};

// a generic plane
struct plane
{
  vec3 n;
  GLfloat d;

  plane () { };
  plane (GLfloat a, GLfloat b, GLfloat c, GLfloat d);
};

inline GLfloat distance (const plane &a, const point &p)
{
  return dot (a.n, p) + a.d;
}

struct sphere {
  point p;
  GLfloat r;

  sphere () { };
  sphere (const point &p, GLfloat r) : p(p), r(r) { };
};

inline bool overlap (const sphere &a, const sphere &b)
{
  GLfloat d = a.r + b.r;

  return length2 (a.p - b.p) <= d * d;
}

struct cone {
  point p;
  vec3 a; // axis
  GLfloat f, fs1, fc2; // angle

  cone () { };
  cone (const point &p, const vec3 &a, GLfloat f)
  : p(p), a(a), f(f)
  {
    fs1 = 1.F / sinf (f);
    fc2 = cosf (f); fc2 *= fc2;
  };
};

inline bool overlap (const cone &c, const sphere &s)
{
  vec3 d = s.p - c.p + c.a * (s.r * c.fs1);
  GLfloat dd = dot (c.a, d);
  return dd > 0.F  && dd * dd >= length2 (d) * c.fc2;
}

void renormalize (sector &s, point &p);

struct colour
{
  GLubyte r, g, b, a;
  colour (GLfloat r = 1., GLfloat g = 1., GLfloat b = 1., GLfloat a = 1.)
  : r(GLubyte (r * 255.F + .5F))
  , g(GLubyte (g * 255.F + .5F))
  , b(GLubyte (b * 255.F + .5F))
  , a(GLubyte (a * 255.F + .5F))
  {
  }
};

struct tex2
{
  GLfloat s, t;
  tex2 () { };
  tex2 (GLfloat s, GLfloat t) : s(s), t(t) { };
};

inline const tex2 operator +(const tex2 &a, const tex2 &b)
{
  return tex2 (a.s + b.s, a.t + b.t);
}

inline const tex2 operator -(const tex2 &a, const tex2 &b)
{
  return tex2 (a.s - b.s, a.t - b.t);
}

inline const tex2 operator *(const tex2 &a, GLfloat s)
{
  return tex2 (a.s * s, a.t * s);
}

template<typename vectype> class vector_traits { };

template<>
struct vector_traits<vec2>
{
  typedef GLfloat basetype;
  static const int dimension () { return 2; }
};

template<>
struct vector_traits<vec3>
{
  typedef GLfloat basetype;
  static const int dimension () { return 3; }
};

template<>
struct vector_traits<vec4>
{
  typedef GLfloat basetype;
  static const int dimension () { return 4; }
};

template<>
struct vector_traits<tex2>
{
  typedef GLfloat basetype;
  static const int dimension () { return 2; }
};

template<>
struct vector_traits<colour>
{
  typedef GLubyte basetype;
  static const int dimension () { return 4; }
};

struct box
{
  point a, b;

  box() { };

  void reset ()
  {
    a = point ( FLT_MAX,  FLT_MAX,  FLT_MAX);
    b = point (-FLT_MAX, -FLT_MAX, -FLT_MAX);
  }

  void add (const box &o);
  void add (const point &p);
};

struct entity;
struct geometry;
struct view;
struct octant;

extern struct timer
{
  static double now;
  static double diff;
  static double fps;

  static void frame ();
  timer ();
} timer;

namespace gl
{
#ifdef DEBUG
  void errchk (const char *name, const char *args, const char *file, int line);
#endif

  struct vertex_v3f
  {
    point v; // vertex

    vertex_v3f () { };
    vertex_v3f (point v) : v(v) { };
  };

  struct vertex_t2f_n3f_v3f
  {
    tex2  t; // texture
    vec3  n; // normal
    point v; // vertex

    vertex_t2f_n3f_v3f () { }
    vertex_t2f_n3f_v3f (point v, vec3 n, tex2 t = tex2()) : v(v), n(n), t(t) { }
  };

  template<typename unused> class type_traits { };

  template<> struct type_traits<GLfloat > { static const GLenum glenum () { return GL_FLOAT; } };
  template<> struct type_traits<GLdouble> { static const GLenum glenum () { return GL_DOUBLE; } };
  template<> struct type_traits<GLint   > { static const GLenum glenum () { return GL_INT; } };
  template<> struct type_traits<GLuint  > { static const GLenum glenum () { return GL_UNSIGNED_INT; } };
  template<> struct type_traits<GLshort > { static const GLenum glenum () { return GL_SHORT; } };
  template<> struct type_traits<GLushort> { static const GLenum glenum () { return GL_UNSIGNED_SHORT; } };
  template<> struct type_traits<GLbyte  > { static const GLenum glenum () { return GL_BYTE; } };
  template<> struct type_traits<GLubyte > { static const GLenum glenum () { return GL_UNSIGNED_BYTE; } };
  template<> struct type_traits<vertex_v3f        > { static const GLenum glenum () { return GL_V3F; } };
  template<> struct type_traits<vertex_t2f_n3f_v3f> { static const GLenum glenum () { return GL_T2F_N3F_V3F; } };

  template<GLenum unused> class enum_traits { };

  template<> struct enum_traits<GL_FLOAT         > { typedef GLfloat type; };
  template<> struct enum_traits<GL_DOUBLE        > { typedef GLdouble type; };
  template<> struct enum_traits<GL_INT           > { typedef GLint type; };
  template<> struct enum_traits<GL_UNSIGNED_INT  > { typedef GLuint type; };
  template<> struct enum_traits<GL_SHORT         > { typedef GLshort type; };
  template<> struct enum_traits<GL_UNSIGNED_SHORT> { typedef GLushort type; };
  template<> struct enum_traits<GL_BYTE          > { typedef GLbyte type; };
  template<> struct enum_traits<GL_UNSIGNED_BYTE > { typedef GLubyte type; };
  template<> struct enum_traits<GL_V3F           > { typedef vertex_v3f type; };
  template<> struct enum_traits<GL_T2F_N3F_V3F   > { typedef vertex_t2f_n3f_v3f type; };

  // deprecated:
  GLsizei format_stride (GLenum format);
  GLsizei format_offset_p (GLenum format);
  GLsizei format_offset_n (GLenum format);
  GLsizei format_offset_t (GLenum format);

  struct matrix
  {
    GLfloat data[4][4];

    const GLfloat operator ()(int i, int j) const { return data[j][i]; };
    GLfloat &operator ()(int i, int j) { return data[j][i]; };

    operator GLfloat *() { return &data[0][0]; }

    void diagonal (GLfloat v);
    void clear () { diagonal (0.F); };
    void identity () { diagonal (1.F); };

    void print (); // ugly

    static matrix translation (const vec3 &v);
    static matrix rotation (GLfloat degrees, const vec3 &axis);
    static matrix scaling (GLfloat sx, GLfloat sy, GLfloat sz, GLfloat w = 1.F);

    matrix () { };
    matrix (GLfloat diag) { diagonal (diag); };
  };

  matrix operator *(const matrix &a, const matrix &b);
  vec3 operator *(const matrix &a, const vec3 &v);

  template<GLenum target>
  struct vertex_buffer_object
  {
    GLuint buffer;
    GLenum format;
    GLsizei count;
    GLsizei size;

    bool alloc ()
    {
      if (buffer)
        return false;

      glGenBuffersARB (1, &buffer);
      return true;
    }

    void bind ()
    {
      glBindBufferARB (target, buffer);
    }

    template<class basetype>
    void set (const basetype *d, GLsizei count, GLenum usage = GL_STATIC_DRAW_ARB)
    {
      alloc ();
      format = type_traits<basetype>::glenum ();
      this->count = count;
      size = sizeof (basetype);
      vertex_buffer_object::bind ();
      glBufferDataARB (target, count * this->size, d, usage);
    }

    template<class data>
    void set (const vector<data> &d, GLenum usage = GL_STATIC_DRAW_ARB)
    {
      set (&d[0], d.size (), usage);
    }

    template<typename basetype>
    void set (GLsizei count, GLenum usage = GL_DYNAMIC_DRAW_ARB)
    {
      alloc ();
      format = type_traits<basetype>::glenum ();
      this->count = count;
      size = sizeof (basetype);
      vertex_buffer_object::bind ();
      glBufferDataARB (target, this->count * size, 0, usage);
    }

    void *map (GLenum access = GL_WRITE_ONLY)
    {
      vertex_buffer_object::bind ();
      return glMapBufferARB (target, access);
    }

    void unmap ()
    {
      vertex_buffer_object::bind ();
      glUnmapBuffer (target);
    }

    vertex_buffer_object ()
    : buffer(0)
    {
    }
    
    ~vertex_buffer_object ()
    {
      if (buffer)
        glDeleteBuffersARB (1, &buffer);
    }

    operator GLint ()
    {
      return buffer;
    }
  };

  struct vertex_buffer : vertex_buffer_object<GL_ARRAY_BUFFER_ARB>
  {
    void bind ()
    {
      vertex_buffer_object<GL_ARRAY_BUFFER_ARB>::bind ();
      glInterleavedArrays (format, 0, 0);
    }

    void draw (GLenum mode, GLint first, GLsizei count)
    {
      bind ();
      glDrawArrays (mode, first, count);
    }
  };

  struct index_buffer : vertex_buffer_object<GL_ELEMENT_ARRAY_BUFFER_ARB>
  {
    void draw (GLenum mode, GLint first, GLsizei count)
    {
      bind ();
      glDrawElements (mode, count, format, (GLvoid *)((char *)0 + first * size)); // only SHORT supported :( // first //D//TODO
    }
  };

  void draw_bbox (const sector &a, const sector &b);

  GLuint load_texture (SDL_Surface *surface, GLfloat *texcoord);
}

template<typename idtype = unsigned int>
struct idpool
{
  typedef idtype type;
  vector<idtype> free;
  idtype next;
 
  idpool ();
  idtype get ();
  void put (idtype &i);
};

template<class value, typename idtype = unsigned int>
struct idmap
{
  vector<value *> map;

  idpool<idtype> &ctx;

  idmap (idpool<idtype> &ctx)
  : ctx (ctx)
  {
  }

  ~idmap ()
  {
  }

#if 0
  value *&operator [](const base &k)
  {
  }
#endif
};

#endif

